// Requires C++ 14
// Compilation: c++ -std=c++14 -Wall -Wno-c++11-extensions human.cc -o human
// Execution: ./human

#include <ctime>

#include <iostream>
#include <memory>
#include <string>

class Human
{
public:
  static void census ();

  Human (const std::string& name, float size, unsigned birth_year);
  ~Human ();

  const std::string& name () const;
  void rename (const std::string& name);

  unsigned age () const;
  void hello () const;

  friend void birth_control (const Human& human);

  // #### NOTE: private is the default status in C++.
private:
  // #### NOTE: initialization outside the class (see below).
  // Exceptions: const or inline (C++17) attributes.
  static unsigned population_;

  std::string name_;
  const float size_;
  const unsigned birth_year_;
};

unsigned Human::population_ = 0;
void Human::census ()
{
  if (population_)
    std::cout << population_ << " human"
	      << (population_ > 1 ? "s " : " ")
	      << "currently alive.\n";
  else
    std::cout << "No human currently alive.\n";
}

Human::Human (const std::string& name, float size, unsigned birth_year)
  : name_ (name), size_ (size), birth_year_ (birth_year)
{
  std::cout << "Creating the human " << name_ << "...\n";
  population_ += 1;
}

Human::~Human ()
{
  std::cout << "Destroying the human " << name_ << "...\n";
  population_ -= 1;
}

const std::string& Human::name () const
{
  return name_;
}

void Human::rename (const std::string& name)
{
  // Maybe check with the administration first ;-)
  name_ = name;
}

unsigned Human::age () const
{
  time_t t = time (0);
  struct tm* tm = localtime (&t);

  return (unsigned) tm->tm_year + 1900 - birth_year_;
}

void Human::hello () const
{
  std::cout << "Hello! I'm " << name_ <<  ", "
	    << size_ << "m, "
	    << age () << "yo.\n";
}

// Sorry, it's very difficult to come up with a simple and reasonable example
// of the friend feature... ;-)
void birth_control (const Human& human)
{
  std::cout << "The NSA knows about " << human.name_ << "...\n";
}

int main ()
{
  Human::census ();
  std::cout << std::endl;

  // Old style construction with explicit constructor call.
  Human h1 = Human ("Alain Térieur", 1.80, 1970);
  h1.census ();
  h1.hello ();
  birth_control (h1);
  std::cout << std::endl;

  // Old style construction with implicit constructor call.
  Human h2 ("Alex Térieur", 1.78, 1975);
  h2.census ();
  h2.hello ();
  birth_control (h2);
  std::cout << std::endl;

  // C++ 11 initializer lists.
  Human h3 { "Vladimir Guez", 1.85, 1980 };
  h3.census ();
  h3.hello ();
  birth_control (h3);
  std::cout << std::endl;

  // Even better C++ 11 style.
  auto h4 = Human { "Anne Titgoutte", 1.85, 1985 };
  h4.census ();
  h4.hello ();
  birth_control (h4);
  std::cout << std::endl;

  // Regular C++ pointers (new / delete).
  Human *h5 = new Human ("Corinne Titgoutte", 1.68, 1990);
  h5->census ();
  h5->hello ();
  birth_control (*h5);
  std::cout << std::endl;

  delete h5;
  Human::census ();
  std::cout << std::endl;

  // Smart pointers (implicit deletion).
  // #### NOTE: make_unique is C++ 14 only.
  auto h6 = std::make_unique<Human> ("Justine Titgoutte", 1.75, 1995);
  h6->census ();
  std::cout << h6->name () << " wants a new name...\n";
  h6->rename ("Anakin Titgoutte");
  h6->hello ();
  birth_control (*h6);
  std::cout << std::endl;
}
